/* Compute correlation functions algebraically on a finite lattice */

/* Written May/June 1996 by Andreas Honecker */
/* E-Mail: honecker@physik.fu-berlin.de */
/* This file may be freely distributed and changed, */
/* provided that your changes are marked clearly */
/* and this and the previous lines remain unchanged */

/* The algortithm used in this program is described mainly
   in appendix C of the paper
   A. Honecker, I. Peschel, "Matrix-Product States for a
   One-Dimensional Lattice Gas with Parallel Dynamics" */
/* Equation numbers refer to our paper */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

/* Values of truth */

#ifndef TRUE
  #define TRUE -1
#endif
#ifndef FALSE
  #define FALSE 0
#endif
#define UNKNOWN 0x55aa55

typedef double REAL;

/* Maximal chain length plus one */

#define MAXSITES 301

/* Chain length */

static int NSITES=100;

/* Parameters */
/* Preset values (seen from outside) */

static REAL alpha_pr = 0.5;
static REAL beta_pr = 0.5;
 
static REAL p_pr = 0.6;
static REAL q_pr = 0.2;
 
static REAL gamma_pr = 0.1666666667;
static REAL delta_pr = 0.3214285714;

/* An undefined probability */

#define PROBUNDEF -4321.0

/* actual values (used internally) */

static REAL alpha, beta, gama, delta, p, q;

/* Density profile */

static REAL density_prof[MAXSITES];

/* Density-density correlation */

static REAL density_corr[MAXSITES][MAXSITES];

/* A few flags */

static int connected_flag = TRUE;
static int only_profile = TRUE;
static int parity = UNKNOWN;
static int reverse_sign = FALSE;

/* Type of the state */

struct state {int dim;
              REAL *coeff;
             };

struct state states[MAXSITES][MAXSITES];

static REAL s_val[MAXSITES];

/* This function is implemented as a Macro */

#define min(a,b) ((a<b)?a:b)
#define max(a,b) ((a<b)?b:a)

extern REAL *access_state(int x, int y);


/* Initialize states */

void init_states()
 {
  register int i, j;

  for(i=0; i<MAXSITES; i++)
   for(j=0; j<MAXSITES; j++)
    {
      states[i][j].coeff = 0;
    }
 }

/* Free memory occupied by states */

void free_states()
 {
  register int i, j;

  for(i=0; i<MAXSITES; i++)
   for(j=0; j<MAXSITES; j++)
    {
      if(states[i][j].coeff)
        free(states[i][j].coeff);
      states[i][j].coeff = 0;
    }
 }

/* Allocate a state */

REAL *alloc_state(int dim)
 {
  REAL *ptr;
  register int i;

  ptr = (REAL *)malloc(sizeof(REAL) * dim);
  if(ptr)
   {
    for(i=0; i<dim; i++)
      ptr[i] = 0;
   }
  else
   {
    fprintf(stderr, "### Error: Out of memory while allocating state\n");
    exit(13);
   }

  return(ptr);
 }

/* Compute a state */

void compute_state(int x, int y)
 {
  register int i;
  REAL *ptr, *ptr1;
  int dim;

  if(y < 0)
   {
    fprintf(stderr, "### Error: index %d is negative\n", y);
    exit(13);
   }

  dim = x+y+2;
  states[x][y].dim = dim;
  ptr = alloc_state(dim);

/*  printf("Creating |%d,%d>\n", x, y); */

  if(! y)            /* use eq. (C.3) for |x,0> */
   {
    ptr[x] = -(1.0-(beta+delta))/delta;
    ptr[x+1] = beta/delta;
   }
  if(y > 0)          /* use eq. (C.2) for general |x,y> */
   {
    ptr[x+y] = 1.0-1.0/q;               /* coefficient of |x+y> in eq. (C.2) */
    ptr1 = access_state(x+1,y-1);
    if(dim != states[x+1][y-1].dim)
     {
      fprintf(stderr, "### Error: Mismatch of dimensions in compute_state()\n");
      exit(13);
     }
    for(i=0; i<dim; i++)
     {
      ptr[i] += (p/q * ptr1[i]);        /* Add p/q |x+1,y-1> */
     }
    ptr1 = access_state(x,y-1);
    if(dim-1 != states[x][y-1].dim)
     {
      fprintf(stderr, "### Error: Mismatch of dimensions in compute_state()\n");
      exit(13);
     }
    for(i=0; i<dim-1; i++)
     {
      ptr[i] += (-(1.0-p)/q * ptr1[i]); /* Subtract (1-p)/q |x,y-1> */
     }
   }
  states[x][y].coeff = ptr;
 }

/* Access a state */

REAL *access_state(int x, int y)
 {
   if((x < 0) || (x >= MAXSITES))
    {
     fprintf(stderr, "### Error: %d outside allowed x range\n", x);
     exit(13);
    }
   if((y < 0) || (y >= MAXSITES))
    {
     fprintf(stderr, "### Error: %d outside allowed y range\n", y);
     exit(13);
    }
   /* Check if the state exists */
   if(states[x][y].coeff)
     return(states[x][y].coeff);
   /* If not, create it */
   compute_state(x,y);
   return(states[x][y].coeff);
 } 

/* Apply the particle operator to a state */
/* (just a shift of the coefficients - see eq. (C.4) ) */

applyB(struct state *src, struct state *dst)
 {
  register int i;

  if(dst->coeff)
    free(dst->coeff);
  dst->dim = src->dim+1;
  dst->coeff = alloc_state(dst->dim + 1);
  for(i=0; i<src->dim; i++)
    dst->coeff[i+1] = src->coeff[i];
 }

/* Apply the particle operator with a hat to a state */
/* (shift plus identity - see eqs. (5.1) and (C.4) ) */

applyBh(struct state *src, struct state *dst)
 {
  register int i;

  if(dst->coeff)
    free(dst->coeff);
  dst->dim = src->dim+1;
  dst->coeff = alloc_state(dst->dim + 1);
  dst->coeff[0] = src->coeff[0];
  for(i=0; i<src->dim; i++)
    dst->coeff[i+1] = src->coeff[i+1] + src->coeff[i];
 }

/* Apply the empty-place operator to a state */

applyA(struct state *src, struct state *dst)
 {
  register int x, i;
  REAL *ptr;

  if(dst->coeff)
    free(dst->coeff);
  dst->dim = src->dim+1;
  dst->coeff = alloc_state(dst->dim + 1);
  for(x=0; x<src->dim; x++)
   {
    ptr = access_state(0,x);            /* Eq. (C.4) : A |x> = |0,x> */
    if(states[0][x].dim != x+2)
     {
      fprintf(stderr, "### Error: Unexpected dimension while applying A\n");
      exit(13);
     }
    for(i=0; i <= x+1; i++)
      dst->coeff[i] += (src->coeff[x] * ptr[i]);
   }
 }

/* Apply the local transfer operator to a state */
/* (C = A + B) */

applyC(struct state *src, struct state *dst)
 {
  struct state tmp1, tmp2;
  register int i;

  if(dst->coeff)
    free(dst->coeff);
  dst->dim = src->dim+1;
  dst->coeff = alloc_state(dst->dim + 1);
  tmp1.coeff = 0;
  tmp2.coeff = 0;
  applyA(src, &tmp1);
  applyB(src, &tmp2);
  if((tmp1.dim != dst->dim) || (tmp2.dim != dst->dim))
   {
    fprintf(stderr, "### Error: Unexpected dimension(s) while applying C\n");
    exit(13);
   }
  for(i=0; i<dst->dim; i++)
    dst->coeff[i] = (tmp1.coeff[i] + tmp2.coeff[i]);
  free(tmp1.coeff);
  free(tmp2.coeff);
 }

/* Compute the parameters characterizing the phase diagram (eqs. (4.1) and (6.1) )*/

REAL kappahpl(REAL x, REAL y, REAL pv, REAL qv)
 {
  REAL xt, yt;
  REAL tt, res;

  xt = x*(1.0-qv);
  yt = y*(1.0-pv);

  tt = -xt + yt + pv - qv;
  res = (tt + sqrt(tt*tt+4.0*xt*yt))/(2.0*x*sqrt((1.0-pv)*(1.0-qv)));
  return(res);
 }

/* This one is needed for the current (eq. (4.1) )*/

REAL kappatmi(REAL x, REAL y, REAL pv, REAL qv)
 {
  REAL xt, yt;
  REAL tt, res;

  xt = x*(1.0-qv);
  yt = y*(1.0-pv);

  tt = -xt + yt + pv - qv;
  res = (tt - sqrt(tt*tt+4.0*xt*yt))/(2.0*x);
  return(res);
 }

/* Read a probability from input unless already specified */

void get_prob(REAL *prob, char *name)
 {
  while(TRUE)
   {
    if(*prob == PROBUNDEF)
     {
      printf("%s = [0..1] -> ", name);
      scanf("%lg", prob);
     }
    if((*prob >= 0.0) && (*prob <= 1.0))
      return;
    printf("\n\n%s is a probability and must lie between 0 and 1 !\n\n", name);
    *prob = PROBUNDEF;
   }
 }

/* This is the main program */

main(int argc, char *argv[])
 {
  register int z, y, x, i;
  REAL div_fac;
  REAL h1, h2;
  struct state CN, HCN, XN, HXN;
  REAL expectC, expectCp;
  REAL expectB;
  REAL ka, kb, ph, qh;
  REAL kat, kbt;
  REAL JtdD, Jtd, JN, densE, densO, expzeta;
  int startx, stepx;
  FILE *fpo;
  char *out_file;
  int exit_flag;

  /* Process command line */

  exit_flag=0;
  out_file = 0;
  connected_flag = FALSE;
  only_profile = FALSE;
  reverse_sign = FALSE;
  parity = UNKNOWN;
  NSITES = UNKNOWN;
  stepx = startx = 2;

  alpha_pr = beta_pr = gamma_pr = delta_pr = p_pr = q_pr = PROBUNDEF;

  out_file = 0;

  for(i=1; i<argc; i++)
    {
     if(argv[i][0] == '-')
      {
       switch(argv[i][1] | 0x20)
        {
         case 'a':
          sscanf(argv[i]+2, "%lg", &alpha_pr);
          break;
         case 'b':
          sscanf(argv[i]+2, "%lg", &beta_pr);
          break;
         case 'g':
          sscanf(argv[i]+2, "%lg", &gamma_pr);
          break;
         case 'd':
          sscanf(argv[i]+2, "%lg", &delta_pr);
          break;
         case 'p':
          sscanf(argv[i]+2, "%lg", &p_pr);
          break;
         case 'q':
          sscanf(argv[i]+2, "%lg", &q_pr);
          break;
         case 'n':
          sscanf(argv[i]+2, "%d", &NSITES);
          break;
         case 'c':
          connected_flag = TRUE;
          break;
         case 'f':
          connected_flag = FALSE;
          break;
         case 'j':
          only_profile = TRUE;
          break;
         case 't':
          only_profile = FALSE;
          break;
         case 'e':
          stepx = startx = 2;
          break;
         case 'o':
          stepx = 2;
          startx = 1;
          break;
         case 'y':
          stepx = 1;
          startx = 1;
          break;
         case 'm':
          reverse_sign = TRUE;
          break;
         case 's':
          parity = FALSE;
          break;
         case 'r':
          parity = TRUE;
          break;
        default:
          exit_flag |= 1;
          break;
        }
      }
     else
      {
       if(! out_file)
         out_file = argv[i];
       else
         exit_flag |= 2;
      }
    }

   if(! out_file)
     exit_flag |= 4;

   if(exit_flag)
    {
     if(exit_flag & 1)
        fprintf(stderr, "**** Illegal option ****\n\n");
     if(exit_flag & 2)
        fprintf(stderr, "**** More than one output file ****\n\n");
     if(exit_flag & 4)
        fprintf(stderr, "**** No output file ****\n\n");
     fprintf(stderr, "USE: %s [-options*] out_file\n", argv[0]);
     fprintf(stderr, "\nPossible options are:\n");
     fprintf(stderr, "-a#\tSet alpha=#\n");
     fprintf(stderr, "-b#\tSet beta=#\n");
     fprintf(stderr, "-g#\tSet gamma=#\n");
     fprintf(stderr, "-d#\tSet delta=#\n");
     fprintf(stderr, "-p#\tSet p=#\n");
     fprintf(stderr, "-q#\tSet q=#\n");
     fprintf(stderr, "-N#\tSet chain length N=#\n");
     fprintf(stderr, "-c\tCompute connected two-point function (<n_x n_y> - <n_x><n_y>)\n");
     fprintf(stderr, "-f\tCompute full two-point function (<n_x n_y>) [default]\n");
     fprintf(stderr, "-j\tCompute just the density-profile <n_x> and no two-point functions\n");
     fprintf(stderr, "-t\tCompute two-point function, profile = <n_x n_x> [default]\n");
     fprintf(stderr, "-e\tCompute quantities only on the even sublattice (n_x) [default]\n");
     fprintf(stderr, "-o\tCompute quantities only on the odd sublattice (\\hat{n}_x)\n");
     fprintf(stderr, "-y\tCompute quantities everywhere on the lattice\n"); 
     fprintf(stderr, "-m\tReverse sign of connected two-point function (use with -c)\n");
     fprintf(stderr, "-r\tUse a spatially reversed system\n");
     fprintf(stderr, "-s\tUse standard order in space\n");
     fprintf(stderr, "\t(if -r/-s is not specified, the decision is taken automatically)\n");
     exit(13);
    }

  /* Get the parameters that are not or incorrectly specified on the command-line */

NSITES_loop:
  if(NSITES == UNKNOWN)
   {
    printf("Number of sites N = [2..%d] -> ", MAXSITES-1);
    scanf("%d", &NSITES);
   }
  if(NSITES < 2)
   {
    printf("\nN must be at least 2 !\n\n");
    NSITES = UNKNOWN;
    goto NSITES_loop;
   }
  if(NSITES >= MAXSITES)
   {
    printf("\nN can be at most %d !\n", MAXSITES-1);
    printf("\tif you really need this size:\n");
    printf("\tchange the definition in the source to\n");
    printf("\t\t#define MAXSITES %d\n", NSITES+1);
    printf("\t... and recompile\n\n");
    NSITES = UNKNOWN;
    goto NSITES_loop;
   }
  if(NSITES & 1)
   {
    printf("\nN must be even !\n");
    printf("(this is included in the definition of our model)\n\n");
    NSITES = UNKNOWN;
    goto NSITES_loop;
   }

  get_prob(&p_pr, "p");
  get_prob(&q_pr, "q");
  get_prob(&alpha_pr, "alpha");
  get_prob(&beta_pr, "beta");
  get_prob(&gamma_pr, "gamma");
  get_prob(&delta_pr, "delta");

  printf("Chain length N = %d\n", NSITES);
  printf("p\t= %1.10g\nq\t= %1.10g\nalpha\t= %1.10g\nbeta\t= %1.10g\ngamma\t= %1.10g\ndelta\t= %1.10g\n",
      p_pr, q_pr, alpha_pr, beta_pr, gamma_pr, delta_pr);
  if(p_pr > q_pr)
   {
     ka = kappahpl(alpha_pr, gamma_pr, p_pr, q_pr);
     kb = kappahpl(beta_pr, delta_pr, p_pr, q_pr);
     ph = p_pr;
     qh = q_pr;
   }
  else
   {
     ka = kappahpl(delta_pr, beta_pr, q_pr, p_pr);
     kb = kappahpl(gamma_pr, alpha_pr, q_pr, p_pr);
     ph = q_pr;
     qh = p_pr;
   }
  kat = ka*sqrt((1.0-qh)*(1.0-ph));
  kbt = kb*sqrt((1.0-qh)*(1.0-ph));
  printf("=> \\hat{\\kappa}_+ at the left boundary: %10.10lg\n", ka);
  printf("   \\hat{\\kappa}_+ at the right boundary: %10.10lg\n\n", kb);
  if((ka < 1.0) && (kb < 1.0))
   {
    printf("   Maximal current phase\n\n");
    /* Compute the current according to eq. (6.7) */
    JtdD = 1.0;
    Jtd =  (sqrt(1.0-qh)-sqrt(1.0-ph))/(sqrt(1.0-qh)+sqrt(1.0-ph));
    printf("   J = %1.10g in the thermodynamic limit\n", Jtd);
   }
  else
   {
    if(ka > kb)
     {
      printf("   Low density phase\n\n");
      /* The densities according to eq. (4.2b) */
      printf("   Expectation for the bulk density: <n_x> = %1.10lg\n", densE = (1.0-ph)/(1.0-ph+kat));
      printf("   Expectation for the bulk density: <\\hat{n}_x> = %1.10lg\n", densO = (1.0-qh)/(1.0-qh+kat));
      /* Compute the current according to eq. (4.5b) */
      if(p_pr > q_pr)
       {
        JtdD = (ph-qh)*(1.0-alpha_pr-gamma_pr);
        if(JtdD)
          Jtd = alpha_pr * ((ph-qh)-(alpha_pr+gamma_pr)*(1.0-qh+kappatmi(alpha_pr, gamma_pr, ph, qh)))/JtdD;
        else
          Jtd = 0.0;
       }
      else
       {
        JtdD = (ph-qh)*(1.0-delta_pr-beta_pr);
        if(JtdD)
          Jtd = delta_pr * ((ph-qh)-(delta_pr+beta_pr)*(1.0-qh+kappatmi(delta_pr, beta_pr, ph, qh)))/JtdD;
        else
          Jtd = 0.0;
       }
      if(JtdD)
        printf("   J = %1.10g in the thermodynamic limit\n", Jtd);
      else
        printf("   Singularity in the formula for J\n");
     }
    else
     {
      printf("   High density phase\n\n");
      /* The densities according to eq. (4.2a) */
      printf("   Expectation for the bulk density: <n_x> = %1.10lg\n", densE = kbt/(1.0-qh+kbt));
      printf("   Expectation for the bulk density: <\\hat{n}_x> = %1.10lg\n", densO = kbt/(1.0-ph+kbt));
      /* Compute the current according to eq. (4.5a) */
      if(p_pr > q_pr)
       {
        JtdD = (ph-qh)*(1.0-beta_pr-delta_pr);
        if(JtdD)
          Jtd = beta_pr * ((ph-qh)-(beta_pr+delta_pr)*(1.0-qh+kappatmi(beta_pr, delta_pr, ph, qh)))/JtdD;
        else
          Jtd = 0.0;
       }
      else
       {
        JtdD = (ph-qh)*(1.0-gamma_pr-alpha_pr);
        if(JtdD)
          Jtd = gamma_pr * ((ph-qh)-(gamma_pr+alpha_pr)*(1.0-qh+kappatmi(gamma_pr, alpha_pr, ph, qh)))/JtdD;
        else
          Jtd = 0.0;
       }
      if(JtdD)
        printf("   J = %1.10g in the thermodynamic limit\n", Jtd);
      else
        printf("   Singularity in the formula for J\n");
     }
    /* The correlation length is given by eq. (4.6) */
    printf("   exp(1/zeta) = %1.10g [if applicable]\n", expzeta = (kat*(1.0-ph+kbt)*(1.0-qh+kbt))/
                                                        (kbt*(1.0-ph+kat)*(1.0-qh+kat)));
   }

  /* Test whether to apply parity */

  h1 = min(alpha_pr, beta_pr);
  h1 = min(h1, p_pr);
  h2 = min(gamma_pr, delta_pr);
  h2 = min(h2, q_pr);
  if(parity == UNKNOWN)
   {
    if(h1 > h2)
      parity = TRUE;
    else
      parity = FALSE;
   }
  if(parity)
   {
    alpha = delta_pr;
    beta = gamma_pr;
    gama= beta_pr;
    delta = alpha_pr;
    p = q_pr;
    q = p_pr;
    /* The counting in space may have to be changed */
    if(stepx == 2)
      startx = 3 - startx;
    printf("--> Working internally with a spatially reversed system\n");
   }
  else
   {
    alpha = alpha_pr;
    beta = beta_pr;
    gama= gamma_pr;
    delta = delta_pr;
    p = p_pr;
    q = q_pr;
   }

  if(max(h1,h2) < 1e-10)
   {
    fprintf(stderr, "### Either {alpha, beta, p} or {gamma, delta, q} must all be non-zero\n");
    exit(13);
   }

  init_states();

  /* Compute the s(x) using eq. (C.5) */

  s_val[0] = 1.0;                       /* s(0) = 1 */
  for(x=1; x<=NSITES; x++)
   {
    access_state(0,x-1);
    if(states[0][x-1].dim != x+1)
      {
        fprintf(stderr, "### Error: Dimension mismatch while computing s(%d)\n", x);
        exit(13);
      }
    div_fac = (1.0 - alpha/gama*states[0][x-1].coeff[x]);
    s_val[x] = -s_val[x-1]/(gama*div_fac);
    for(i=0; i<x; i++)
      s_val[x] += alpha/(gama*div_fac)*states[0][x-1].coeff[i]*s_val[i];
    printf("s(%d) = %10.10lg\n", x, s_val[x]);
   }

  CN.dim = 1;
  CN.coeff = alloc_state(1);
  CN.coeff[0] = 1.0;
  expectCp = 1.0;
  /* Compute xth power of C on |V> */
  for(x=1; x <= NSITES; x++)
   {
    HCN.coeff = 0;
    applyC(&CN, &HCN);
    free(CN.coeff);
    CN.dim = HCN.dim;
    CN.coeff = HCN.coeff;
    /* Compute <W| C^x |V> - will be needed for the current */
    if(CN.dim != x+1)
     {
       fprintf(stderr, "### Error: unexpected dimension of C^%d |V>\n", x);
       exit(13);
     }
    expectC = 0;
    for(y=x; y >= 0; y--)
      {
       expectC += (CN.coeff[y] * s_val[y]);
      }
    /* The current is given by eq. (6.2) */
    printf("J(p=%g,q=%g) = <W| C^%d |V> / <W| C^%d |V> = %10.10lg\n", p_pr, q_pr, x-1, x,
                         JN = parity?-expectCp/expectC:expectCp/expectC);
    expectCp = expectC;
   }

  printf("<W| C^%d |V> = %10.10lg\n", NSITES, expectC);
  free(CN.coeff);

  /* Now compute density profile */

  printf("Now compute the density profile ...\n");

  for(y=startx; y<=NSITES; y += stepx)
   {
    XN.dim = 1;
    XN.coeff = alloc_state(1);
    XN.coeff[0] = 1.0;
    /* Compute xth power of C or B on |V> */
    for(x=1; x <= NSITES; x++)
     {
      HXN.coeff = 0;
      if(x != y)
        applyC(&XN, &HXN);
      else
       {
        if(x & 1)
          applyBh(&XN, &HXN);
        else
          applyB(&XN, &HXN);
       }
      free(XN.coeff);
      XN.dim = HXN.dim;
      XN.coeff = HXN.coeff;
     }
 
    if(XN.dim != NSITES+1)
     {
       fprintf(stderr, "### Error: unexpected dimension of C^%d B C^%d |V>\n", NSITES-y, y-1);
       exit(13);
     }
 
    expectB = 0;
    for(x=0; x <= NSITES; x++)
       expectB += (XN.coeff[x] * s_val[x]);
 
    density_prof[y] = expectB/expectC;
    free(XN.coeff);
   }

 /* ... and compute density-density correlations */

 if(! only_profile)
  {
  printf("... and compute density-density correlations\n");
  for(y=startx; y<=NSITES; y += stepx)
   {
   for(z=y; z<=NSITES; z += stepx)
   {
    XN.dim = 1;
    XN.coeff = alloc_state(1);
    XN.coeff[0] = 1.0;
    /* Compute xth power of C or B on |V> */
    for(x=1; x <= NSITES; x++)
     {
      HXN.coeff = 0;
      if((x != y) && (x != z))
        applyC(&XN, &HXN);
      else
       {
        if(x & 1)
          applyBh(&XN, &HXN);
        else
          applyB(&XN, &HXN);
       }
      free(XN.coeff);
      XN.dim = HXN.dim;
      XN.coeff = HXN.coeff;
     }

    if(XN.dim != NSITES+1)
     {
       fprintf(stderr, "### Error: unexpected dimension of C^%d B C^%d |V>\n", NSITES-y, y-1);
       exit(13);
     }

    expectB = 0;
    for(x=0; x <= NSITES; x++)
       expectB += (XN.coeff[x] * s_val[x]);

    density_corr[y][z] = expectB/expectC;
    free(XN.coeff);
   }
  }
  }

  /* Now write result of computation to output file */

  fpo = fopen(out_file, "w");
  if(! fpo)
   {
    fprintf(stderr, "### Error: Failed to open '%s' for writing\n", out_file);
    exit(13);
   }

  printf("... writing the results to '%s'\n", out_file);

  /* Write parameters used */

  fprintf(fpo, "# Sytem size N = %d\n", NSITES);
  fprintf(fpo, "# p = %1.10g, q = %1.10g, alpha = %1.10g, beta = %1.10g, gamma = %1.10g, delta = %1.10g\n",
      p_pr, q_pr, alpha_pr, beta_pr, gamma_pr, delta_pr);
  fprintf(fpo, "#\n");

  /* And now write what this implies according to our knowledge */

  fprintf(fpo, "# => \\hat{\\kappa}_+ at the left boundary: %10.10lg\n", ka);
  fprintf(fpo, "#    \\hat{\\kappa}_+ at the right boundary: %10.10lg\n#\n", kb);
  if((ka < 1.0) && (kb < 1.0))
   {
    fprintf(fpo, "#    Maximal current phase\n#\n");
   }
  else
   {
    if(ka > kb)
      fprintf(fpo, "#    Low density phase\n");
    else
      fprintf(fpo, "#    High density phase\n");
    fprintf(fpo, "#\n#    Expectation for the bulk density: <n_x> = %1.10lg\n", densE);
    fprintf(fpo, "#    Expectation for the bulk density: <\\hat{n}_x> = %1.10lg\n", densO);
    fprintf(fpo, "#    exp(1/zeta) = %1.10g [if applicable]\n", expzeta);
   }
  if(JtdD)
    fprintf(fpo, "#    J = %1.10g in the thermodynamic limit\n#\n", Jtd);
  else
    fprintf(fpo, "#    Singularity in the formula for J in the thermodynamic limit\n#\n");
  fprintf(fpo, "#    J = %1.10g (result on the lattice with N = %d)\n#\n", JN, NSITES);

  /* Density profile */

  if(only_profile)
   {
    fprintf(fpo, "# <n_x>\n");
    for(y=startx; y<=NSITES; y += stepx)
     {
      fprintf(fpo, "%d\t%10.10lg\n",  parity?y-1:NSITES-y, density_prof[y]);
     }
   }
  else
   {
  if(! connected_flag)
   {
    fprintf(fpo, "# <n_x>\n");
    for(y=startx; y<=NSITES; y += stepx)
     {
      fprintf(fpo, "%d\t%d\t%10.10lg\n",  parity?y-1:NSITES-y, parity?y-1:NSITES-y, density_prof[y]);
     }
    fprintf(fpo, "\n# <n_x n_y>\n");
   }
  else
   fprintf(fpo, "# <n_x n_y> - <n_x> <n_y>\n");

  if(! connected_flag)
    reverse_sign = FALSE;

  /* Correlations first pass */

  for(y=startx; y<=NSITES; y += stepx)
   {
    for(z=y; z<=NSITES; z += stepx)
     if((! connected_flag) || (z != y))
     {
     fprintf(fpo, "%d\t%d\t%10.10lg\n",  parity?z-1:NSITES-y, parity?y-1:NSITES-z, (reverse_sign?(-1.0):(1.0))*
             (density_corr[y][z]-(connected_flag?density_prof[y]*density_prof[z]:0)));
    }
    fprintf(fpo, "\n");
   }

  /* Correlations second pass */

  for(z=startx; z<=NSITES; z += stepx)
   {
    for(y=startx; y<=z; y += stepx)
     if((! connected_flag) || (z != y))
    {
     fprintf(fpo, "%d\t%d\t%10.10lg\n",  parity?z-1:NSITES-y, parity?y-1:NSITES-z, (reverse_sign?(-1.0):(1.0))*
             (density_corr[y][z]-(connected_flag?density_prof[y]*density_prof[z]:0)));
    }
    fprintf(fpo, "\n");
   }
   } /* End of else of if(! only_profile) */

  fclose(fpo);

  free_states();
 }
